import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Method;

import candidates.Blake;
import candidates.BlueMidnightWish;
import candidates.CubeHash;
import candidates.Echo;
import candidates.Fugue;
import candidates.Groestl;
import candidates.Hamsi;
import candidates.JH;
import candidates.Keccak;
import candidates.Luffa;
import candidates.SHAvite3;
import candidates.SIMD;
import candidates.Shabal;
import candidates.Skein;


@SuppressWarnings("unchecked")
public class CheckKAT {

	static long databitlen = 0;
	static byte[] data = null;

	public static void main(String[] args) {

		Class[] algorithms = { SHAvite3.class, Blake.class, BlueMidnightWish.class, Shabal.class, Skein.class, 
				Luffa.class, Hamsi.class, Keccak.class, CubeHash.class, SIMD.class, Groestl.class, Echo.class,
				JH.class, Fugue.class };
		start(algorithms);

	}

	private static void start(Class[] algorithms) {

		for (Class algorithm : algorithms) {
			System.out.println("Now testing: " + algorithm.getSimpleName());
			boolean result = performCheck(algorithm);
			System.out.println(result ? "100% Match" : "");
			System.out.println(" ");
		}
		System.out.println("All done with "+algorithms.length+" algoritms");

	}

	public static boolean performCheck(Class clazz) {

		File file = new File(System.getProperty("user.dir") + "/shortmsgkat/"
				+ clazz.getSimpleName() + "ShortMsgKAT_256.txt");

		Method method = null;
		try {
			method = clazz.getDeclaredMethod("Hash", new Class[] {
					byte[].class, long.class, byte[].class });
		} catch (Exception e1) {
			System.out.println("Unable to create reflection method");
		}

		try {
			BufferedReader br = new BufferedReader(new InputStreamReader(
					new FileInputStream(file)));
			String line;

			while ((line = br.readLine()) != null) {

				if (line.startsWith("Len")) {
					databitlen = Long.parseLong((line.split("= ")[1]));
				}
				if (line.startsWith("Msg")) {
					String value = null;
					int pointer = 0;
					String tmp = line.split("= ")[1];
					data = new byte[tmp.length() / 2];
					for (int i = 0; i < data.length; i++) {
						value = tmp.substring(pointer, pointer + 2);
						data[i] = (byte) Integer.parseInt(value.trim(), 16);
						pointer += 2;
					}
				}

				if (line.startsWith("MD")) {
					byte[] hashval = new byte[32];
					try {
						method.invoke("foo", data, databitlen, hashval);
					} catch (Exception e) {
						System.err
								.println("Algorithm threw an Exception. Stacktrace is commented out\n");
						//e.printStackTrace();
					}

					StringBuffer sb = new StringBuffer(hashval.length * 2);
					for (int i = 0; i < 32; ++i) {
						sb.append(hexChar[(hashval[i] & 0xf0) >>> 4]);
						sb.append(hexChar[hashval[i] & 0x0f]);

					}

					String tmp = line.split("= ")[1];

					if (!tmp.equals(sb.toString())) {
						error(databitlen, sb.toString(), tmp);
						return false;
					}
				}

			}
			br.close();
		} catch (IOException e) {
			System.err.println(clazz.getSimpleName()
					+ "ShortMsgKAT_256 not found");
			return false;
		}

		return true;

	}

	private static void error(long databitlen, String hashval, String tmp) {
		System.out.println("Got an error for len: " + databitlen);
		System.out.println("Your hash:  " + hashval);
		System.out.println("Correct is: " + tmp);
	}

	static char[] hexChar = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
			'A', 'B', 'C', 'D', 'E', 'F' };
}
