public class SIMD {

	private static int[] A = new int[4];
	private static int[] B = new int[4];
	private static int[] C = new int[4];
	private static int[] D = new int[4];
	private static byte[] buffer;

	private static int blocksize, n_feistels;
	private static long count;
	private static int dataPointer;

	private static int p4[][] = { { 1, 0, 3, 2 }, { 2, 3, 0, 1 }, { 3, 2, 1, 0 } };

	private static final int[] P = { 4, 6, 0, 2, 7, 5, 3, 1, 15, 11, 12, 8, 9, 13, 10, 14, 17, 18, 23, 20, 22, 21,
			16, 19, 30, 24, 25, 31, 27, 29, 28, 26 };

	private static final int[] IV_256 = { 0x4d567983, 0x07190ba9, 0x8474577b, 0x39d726e9, 0xaaf3d925, 0x3ee20b03,
			0xafd5e751, 0xc96006d3, 0xc2c2ba14, 0x49b3bcb4, 0xf67caf46, 0x668626c9, 0xe2eaa8d2, 0x1ff47833,
			0xd0c661a5, 0x55693de1 };

	private static int PACK(byte[] M, int i) {
		return (M[i] & 255) ^ ((M[i + 1] & 255) << 8) ^ ((M[i + 2] & 255) << 16) ^ ((M[i + 3] & 255) << 24);
	}

	private static int RequiredAlignment() {
		return 1;
	}

	private static int IF(final int x, final int y, final int z) {
		return ((x & y) | (~x & z));
	}

	private static int MAJ(final int x, final int y, final int z) {
		return ((z & y) | ((z | y) & x));
	}

	private static int T32(int x) {
		return ((x) & 0xFFFFFFFF);
	}

	private static int ROTL32(int x, int n) {
		return T32(((x) << (n)) | ((x) >>> (32 - (n))));
	}

	public static void Init() {
		buffer = new byte[80];
		InitIV(IV_256);
	}

	private static void InitIV(int[] IV) {

		n_feistels = 4;
		blocksize = 512;
		count = 0;

		System.arraycopy(IV, 0, A, 0, 4);
		System.arraycopy(IV, 4, B, 0, 4);
		System.arraycopy(IV, 8, C, 0, 4);
		System.arraycopy(IV, 12, D, 0, 4);

	}

	public static void Update(byte[] data, long databitlen) {

		dataPointer = data.length;
		int current;
		int bs = blocksize;
		int align = -1;

		if (align == -1)
			align = RequiredAlignment();
		current = (int) (count & (bs - 1));

		while (databitlen > 0) {
			if (current == 0 && databitlen >= bs) {

				SIMD_Compress(data, 0);
				databitlen -= bs;
				System.arraycopy(data, bs / 8, data, 0, dataPointer - bs / 8);
				for (int i = 0; i < bs / 8; i++) {
					data[i + dataPointer - bs / 8] = 0;
				}
				dataPointer -= bs / 8;
				count += bs;
			} else {

				int len = bs - current;
				if (databitlen < len) {
					System.arraycopy(data, 0, buffer, current / 8, (int) ((databitlen + 7) / 8));
					count += databitlen;
					return;
				} else {

					System.arraycopy(data, 0, buffer, current / 8, len / 8);
					count += len;
					databitlen -= len;
					System.arraycopy(data, len / 8, data, 0, dataPointer - len / 8);
					for (int i = 0; i < data.length; i++) {
						data[i + dataPointer - len] = 0;
					}
					dataPointer -= len / 8;
					current = 0;
					SIMD_Compress(buffer, 0);
				}
			}
		}

	}

	public static void Final(byte[] hashval) {

		long l;
		int current = (int) (count & (blocksize - 1));
		byte[] bs = new byte[64];
		int isshort = 1;

		if (current != 0) {
			if ((current & 7) != 0) {
				byte mask = (byte) (0xff >> (current & 7));
				buffer[current / 8] &= ~mask;
			}
			current = (current + 7) / 8;
			SIMD_Compress(buffer, 0);
		}

		for (int j = 0; j < 64; j++) {
			buffer[j] = 0;
		}

		l = count;
		for (int i = 0; i < 8; i++) {
			buffer[i] = (byte) (l & 0xff);
			l >>= 8;
		}
		if (count < 16384)
			isshort = 2;

		SIMD_Compress(buffer, isshort);

		for (int i = 0; i < n_feistels; i++) {
			int x = A[i];
			bs[4 * i] = (byte) (x & 0xff);
			x >>= 8;
			bs[4 * i + 1] = (byte) (x & 0xff);
			x >>= 8;
			bs[4 * i + 2] = (byte) (x & 0xff);
			x >>= 8;
			bs[4 * i + 3] = (byte) (x & 0xff);
		}
		for (int i = 4; i < 2 * n_feistels; i++) {
			int x = B[i - 4];
			bs[4 * i] = (byte) (x & 0xff);
			x >>= 8;
			bs[4 * i + 1] = (byte) (x & 0xff);
			x >>= 8;
			bs[4 * i + 2] = (byte) (x & 0xff);
			x >>= 8;
			bs[4 * i + 3] = (byte) (x & 0xff);
		}

		System.arraycopy(bs, 0, hashval, 0, 32);

	}

	private static void SIMD_Compress(byte[] M, int finale) {

		int[][] W = new int[32][8];
		int[][] IV = new int[4][8];

		for (int i = 0; i < n_feistels; i++) {
			IV[0][i] = A[i];
			IV[1][i] = B[i];
			IV[2][i] = C[i];
			IV[3][i] = D[i];
		}
		message_expansion(W, M, finale);
		for (int j = 0; j < n_feistels; j++) {
			A[j] ^= PACK(M, 4 * j);
			B[j] ^= PACK(M, 4 * j + 4 * n_feistels);
			C[j] ^= PACK(M, 4 * j + 8 * n_feistels);
			D[j] ^= PACK(M, 4 * j + 12 * n_feistels);
		}

		Round(W, 0, 3, 23, 17, 27);
		Round(W, 1, 28, 19, 22, 7);
		Round(W, 2, 29, 9, 15, 5);
		Round(W, 3, 4, 13, 10, 25);

		Step(IV[0], 32, 4, 13, 0);
		Step(IV[1], 33, 13, 10, 0);
		Step(IV[2], 34, 10, 25, 0);
		Step(IV[3], 35, 25, 4, 0);

	}

	private static void Round(int[][] w, final int i, final int r, final int s, final int t, final int u) {
		
		Step(w[8 * i], 8 * i, r, s, 0);
		Step(w[8 * i + 1], 8 * i + 1, s, t, 0);
		Step(w[8 * i + 2], 8 * i + 2, t, u, 0);
		Step(w[8 * i + 3], 8 * i + 3, u, r, 0);

		Step(w[8 * i + 4], 8 * i + 4, r, s, 1);
		Step(w[8 * i + 5], 8 * i + 5, s, t, 1);
		Step(w[8 * i + 6], 8 * i + 6, t, u, 1);
		Step(w[8 * i + 7], 8 * i + 7, u, r, 1);
	}

	private static void Step(int[] w, final int i, final int r, final int s, int boolean_function) {

		int[] tmp = new int[8];
		int[] perm = new int[4];

		for (int j = 0; j < n_feistels; j++)
			tmp[j] = ROTL32(A[j], r);

		for (int k = 0; k < 4; k++)
			perm[k] = p4[i % 3][k];

		for (int j = 0; j < n_feistels; j++) {
			int p = perm[j];
			A[j] = D[j] + w[j] + (boolean_function == 0 ? IF(A[j], B[j], C[j]) : MAJ(A[j], B[j], C[j]));
			A[j] = T32(ROTL32(T32(A[j]), s) + tmp[p]);
			D[j] = C[j];
			C[j] = B[j];
			B[j] = tmp[j];
		}

	}

	private static void message_expansion(int[][] W, byte[] M, int finale) {

		int[] y = new int[256];
		final int alpha = 139;

		int beta = 98;
		int beta_i = 1;
		int alpha_i = 1;
		int[][] Z = new int[32][8];

		final int fft_size = 128;
		final int M_size = 64;

		for (int i = 0; i < fft_size; i++) {
			y[i] = beta_i;
			beta_i = (beta_i * beta) % 257;
		}

		if (finale != 0) {
			beta = 58;
			beta_i = 1;
			for (int i = 0; i < fft_size; i++) {
				y[i] += beta_i;
				beta_i = (beta_i * beta) % 257;
			}
		}

		for (int i = 0; i < fft_size; i++) {
			int alpha_ij = 1;

			for (int j = 0; j < M_size; j++) {

				y[i] = (y[i] + alpha_ij * (M[j] & 255)) % 257;
				alpha_ij = (alpha_ij * alpha_i) % 257;
			}

			alpha_i = (alpha_i * alpha) % 257;
		}

		for (int i = 0; i < fft_size; i++)
			if (y[i] > 128)
				y[i] -= 257;

		for (int i = 0; i < 16; i++)
			for (int j = 0; j < n_feistels; j++)
				Z[i][j] = (((int) (y[2 * i * n_feistels + 2 * j] * 185)) & 0xffff)
						| ((int) (y[2 * i * n_feistels + 2 * j + 1] * 185) << 16);

		for (int i = 0; i < 8; i++)
			for (int j = 0; j < n_feistels; j++)
				Z[i + 16][j] = (((int) (y[2 * i * n_feistels + 2 * j] * 233)) & 0xffff)
						| ((int) (y[2 * i * n_feistels + 2 * j + fft_size / 2] * 233) << 16);

		for (int i = 0; i < 8; i++)
			for (int j = 0; j < n_feistels; j++)
				Z[i + 24][j] = (((int) (y[2 * i * n_feistels + 2 * j + 1] * 233)) & 0xffff)
						| ((int) (y[2 * i * n_feistels + 2 * j + fft_size / 2 + 1] * 233) << 16);

		for (int i = 0; i < 32; i++)
			for (int j = 0; j < n_feistels; j++)
				W[i][j] = Z[P[i]][j];
	}

	public static void Hash(byte[] data, long databitlen, byte[] hashval) {

		Init();
		Update(data, databitlen);
		Final(hashval);

	}

}
