public class Skein {

	private static final int CFG_FINAL = 0, OUT_FINAL = 1, CFG = 2, KEY = 3, PERS = 4, PK = 5, KDF = 6,
			NONCE = 7, MSG = 8, OUT = 9, MASK = 10;
	private static byte[] cfgb = new byte[32];
	private static long[] cfgw = new long[4];
	private static long[] T = new long[2];
	private static byte[] b = new byte[32];
	private static int bCnt;
	private static long[] ctxX;

	private static long SKEIN_T1_FLAG_FIRST() {
		return (((long) 1) << SKEIN_T1_POS_FIRST());
	}

	private static long SKEIN_T1_FLAG_FINAL() {
		return (((long) 1) << SKEIN_T1_POS_FINAL());
	}

	private static long SKEIN_T1_FLAG_BIT_PAD() {
		return ((long) 1) << SKEIN_T1_POS_BIT_PAD();
	}

	private static long SKEIN_SCHEMA_VER() {
		return (0x33414853 + (((long) 1) << 32));
	}

	private static long SKEIN_KS_PARITY() {
		return ((0x55555555) + (((long) (0x55555555)) << 32));
	}

	private static long SKEIN_CFG_TREE_INFO(int leaf, int node, int maxLvl) {
		return ((((long) (leaf)) << 0) | (((long) (node)) << 8) | (((long) (maxLvl)) << 16));
	}

	private static long SKEIN_CFG_TREE_INFO_SEQUENTIAL() {
		return SKEIN_CFG_TREE_INFO(0, 0, 0);
	}

	private static long SKEIN_T1_BIT(long BIT) {
		return ((BIT) - 64);
	}

	private static long SKEIN_T1_POS_BIT_PAD() {
		return SKEIN_T1_BIT(119);
	}

	private static long SKEIN_T1_POS_BLK_TYPE() {
		return SKEIN_T1_BIT(120);
	}

	private static long SKEIN_T1_POS_FIRST() {
		return SKEIN_T1_BIT(126);
	}

	private static long SKEIN_T1_POS_FINAL() {
		return SKEIN_T1_BIT(127);
	}

	private static long RotL_64(long x, int N) {
		return ((x << (N & 63)) | (x >>> ((64 - N) & 63)));
	}

	private static void InjectKey(int r, int WCNT, long[] X, long[] ks, long[] ts) {
		for (int i = 0; i < WCNT; i++)
			X[i] += ks[((r) + i) % (WCNT + 1)];
		X[WCNT - 3] += ts[((r) + 0) % 3];
		X[WCNT - 2] += ts[((r) + 1) % 3];
		X[WCNT - 1] += (r);
	}

	private static void Skein_Clear_First_Flag() {
		T[1] &= ~SKEIN_T1_FLAG_FIRST();
	}

	private static void Skein_Set_Tweak(int TWK_NUM, long l) {
		T[TWK_NUM] = l;
	}

	private static void Skein_Set_T0(long T0) {
		Skein_Set_Tweak(0, T0);
	}

	private static void Skein_Set_T1(long l) {
		Skein_Set_Tweak(1, l);
	}

	private static void Skein_Set_T0_T1(long T0, long T1) {
		Skein_Set_T0(T0);
		Skein_Set_T1(T1);
	}

	public static void Init() {

		ctxX = new long[4];
		Skein_Start_New_Type(CFG_FINAL);
		cfgw[0] = SKEIN_SCHEMA_VER();
		cfgw[1] = 256;
		cfgw[2] = SKEIN_CFG_TREE_INFO_SEQUENTIAL();
		syncFromLong();
		Skein_256_Process_Block(cfgb, 1, 32);
		Skein_Start_New_Type(MSG);

	}

	public static void Update(byte[] data, long databitlen) {

		if ((databitlen & 7) == 0)
			Skein_256_Update(data, databitlen >>> 3);
		else {
			int bCnt = (int) ((databitlen >>> 3) + 1);
			byte[] b = new byte[1];
			byte mask;
	
			mask = (byte) (1 << (7 - (databitlen & 7)));
			b[0] = (byte) ((data[(bCnt - 1)] & (0 - mask)) | mask);
			Skein_256_Update(data, bCnt - 1);
			Skein_256_Update(b, 1);
			T[1] |= SKEIN_T1_FLAG_BIT_PAD();
		}
	}

	private static void Skein_256_Update(byte[] msg, long msgByteCnt) {
		long n;

		if (msgByteCnt + bCnt > 32) {
			if (bCnt != 0) {
				n = 32 - bCnt;
				if (n != 0) {
					System.arraycopy(msg, 0, b, bCnt, (int) n);
					msgByteCnt -= n;
					System.arraycopy(msg, (int) n, msg, 0, (int) (msg.length - n));
					for (int i = 0; i < n; i++) {
						msg[(int) (i + (msg.length - n))] = 0;
					}
					bCnt += n;
				}
				Skein_256_Process_Block(b, 1, 32);
				bCnt = 0;
			}
			if (msgByteCnt > 32) {
				n = (msgByteCnt - 1) / 32;
				Skein_256_Process_Block(msg, n, 32);
				msgByteCnt -= n * 32;
				System.arraycopy(msg, (int) n * 32, msg, 0, (int) (msg.length - n * 32));
				for (int i = 0; i < n * 32; i++) {
					msg[(int) (i + msg.length - n * 32)] = 0;
				}
			}
		}

		if (msgByteCnt != 0) {
			System.arraycopy(msg, 0, b, bCnt, (int) msgByteCnt);
			bCnt += msgByteCnt;
		}

	}

	private static void Skein_256_Process_Block(byte[] blkPtr, long blkCnt, int byteCntAdd) {

		int offset = 0;
		int WCNT = 4;

		long[] ts = new long[3];
		long[] ks = new long[WCNT + 1];
		long[] X = new long[WCNT];
		long[] w = new long[WCNT];

		while (blkCnt-- != 0) {
			T[0] += byteCntAdd;
			ks[WCNT] = SKEIN_KS_PARITY();
			for (int i = 0; i < WCNT; i++) {
				ks[i] = ctxX[i];
				ks[WCNT] ^= ctxX[i];
			}
			ts[0] = T[0];
			ts[1] = T[1];
			ts[2] = ts[0] ^ ts[1];
			Skein_Get64_LSB_First(blkPtr, offset, w, WCNT);
			for (int i = 0; i < WCNT; i++)
				X[i] = (w[i] + ks[i]);
			X[WCNT - 3] += ts[0];
			X[WCNT - 2] += ts[1];
			for (int r = 1; r <= 72 / 8; r++) {
				X[0] += X[1];
				X[1] = RotL_64(X[1], 14);
				X[1] ^= X[0];
				X[2] += X[3];
				X[3] = RotL_64(X[3], 16);
				X[3] ^= X[2];

				X[0] += X[3];
				X[3] = RotL_64(X[3], 52);
				X[3] ^= X[0];
				X[2] += X[1];
				X[1] = RotL_64(X[1], 57);
				X[1] ^= X[2];

				X[0] += X[1];
				X[1] = RotL_64(X[1], 23);
				X[1] ^= X[0];
				X[2] += X[3];
				X[3] = RotL_64(X[3], 40);
				X[3] ^= X[2];

				X[0] += X[3];
				X[3] = RotL_64(X[3], 5);
				X[3] ^= X[0];
				X[2] += X[1];
				X[1] = RotL_64(X[1], 37);
				X[1] ^= X[2];

				InjectKey(2 * r - 1, WCNT, X, ks, ts);

				X[0] += X[1];
				X[1] = RotL_64(X[1], 25);
				X[1] ^= X[0];
				X[2] += X[3];
				X[3] = RotL_64(X[3], 33);
				X[3] ^= X[2];

				X[0] += X[3];
				X[3] = RotL_64(X[3], 46);
				X[3] ^= X[0];
				X[2] += X[1];
				X[1] = RotL_64(X[1], 12);
				X[1] ^= X[2];

				X[0] += X[1];
				X[1] = RotL_64(X[1], 58);
				X[1] ^= X[0];
				X[2] += X[3];
				X[3] = RotL_64(X[3], 22);
				X[3] ^= X[2];

				X[0] += X[3];
				X[3] = RotL_64(X[3], 32);
				X[3] ^= X[0];
				X[2] += X[1];
				X[1] = RotL_64(X[1], 32);
				X[1] ^= X[2];
				InjectKey(2 * r, WCNT, X, ks, ts);
			}

			for (int i = 0; i < WCNT; i++)
				ctxX[i] = (X[i] ^ w[i]);
			Skein_Clear_First_Flag();
			offset += 32;

		}

	}

	private static void Skein_Get64_LSB_First(byte[] src, int offset, long[] dst, int wCnt) {
		int n;

		for (n = 0; n < 8 * wCnt; n += 8)
			dst[n / 8] = (((long) src[n + offset] & 255)) + (((long) src[n + 1 + offset] & 255) << 8)
					+ (((long) src[n + 2 + offset] & 255) << 16) + (((long) src[n + 3 + offset] & 255) << 24)
					+ (((long) src[n + 4 + offset] & 255) << 32) + (((long) src[n + 5 + offset] & 255) << 40)
					+ (((long) src[n + 6 + offset] & 255) << 48) + (((long) src[n + 7 + offset] & 255) << 56);
	}

	private static void Skein_Start_New_Type(int blk_type) {
		Skein_Set_T0_T1(0, SKEIN_T1_FLAG_FIRST() | SKEIN_T1_BLK_TYPE_(blk_type));
		bCnt = 0;
	}

	private static long SKEIN_T1_BLK_TYPE_(int blk_type) {
		switch (blk_type) {
		case CFG_FINAL:
			return (SKEIN_T1_BLK_TYPE_(CFG) | SKEIN_T1_FLAG_FINAL());
		case OUT_FINAL:
			return (SKEIN_T1_BLK_TYPE_(OUT) | SKEIN_T1_FLAG_FINAL());
		case CFG:
			return (long) 4 << SKEIN_T1_POS_BLK_TYPE();
		case KEY:
			return (long) 0 << SKEIN_T1_POS_BLK_TYPE();
		case PERS:
			return (long) 8 << SKEIN_T1_POS_BLK_TYPE();
		case PK:
			return (long) 12 << SKEIN_T1_POS_BLK_TYPE();
		case KDF:
			return (long) 16 << SKEIN_T1_POS_BLK_TYPE();
		case NONCE:
			return (long) 20 << SKEIN_T1_POS_BLK_TYPE();
		case MSG:
			return (long) 48 << SKEIN_T1_POS_BLK_TYPE();
		case OUT:
			return (long) 63 << SKEIN_T1_POS_BLK_TYPE();
		case MASK:
			return (long) 63 << SKEIN_T1_POS_BLK_TYPE();

		default:
			break;
		}
		return 0;
	}

	public static void Final(byte[] hashVal) {

		int n, byteCnt, offset;
		long[] X = new long[4];

		T[1] |= SKEIN_T1_FLAG_FINAL();
		if (bCnt < 32) {
			for (int i = 0; i < 32 - bCnt; i++) {
				b[i + bCnt] = 0;
			}
		}
		Skein_256_Process_Block(b, 1, bCnt);
		byteCnt = 263 >> 3;
		for (int i = 0; i < b.length; i++)
			b[i] = 0;
		System.arraycopy(ctxX, 0, X, 0, X.length);
		for (int i = 0; i * 32 < byteCnt; i++) {
			b[0] = (byte) ((long) i);
			Skein_Start_New_Type(OUT_FINAL);
			Skein_256_Process_Block(b, 1, 0x8);
			n = byteCnt - i * 32;
			if (n >= 32)
				n = 32;
			offset = i * 32;
			Skein_Put64_LSB_First(hashVal, offset, n);
			System.arraycopy(X, 0, ctxX, 0, X.length);
		}

	}

	private static void Skein_Put64_LSB_First(byte[] dst, int offset, int bCnt) {
		for (int n = 0; n < bCnt; n++)
			dst[n + offset] = (byte) (ctxX[(n) >>> 3] >>> (8 * ((n) & 7)));
	}

	public static void Hash(byte[] data, long databitlen, byte[] hashval) {

		Init();
		Update(data, databitlen);
		Final(hashval);

	}

	private static void syncFromLong() {
		int a = 0;
		for (int i = 0; i < 4; i++) {
			cfgb[a++] = (byte) (255 & cfgw[i]);
			cfgb[a++] = (byte) (255 & cfgw[i] >>> 8);
			cfgb[a++] = (byte) (255 & cfgw[i] >>> 16);
			cfgb[a++] = (byte) (255 & cfgw[i] >>> 24);
			cfgb[a++] = (byte) (255 & cfgw[i] >>> 32);
			cfgb[a++] = (byte) (255 & cfgw[i] >>> 40);
			cfgb[a++] = (byte) (255 & cfgw[i] >>> 48);
			cfgb[a++] = (byte) (255 & cfgw[i] >>> 56);
		}

	}

}
